--  title: ME^2: Matt's Euphoria Mandelbrot Explorer
--     by: Matthew Sephton, Tektonix Software 
-- e-mail: u5ms@csc.liv.ac.uk
--    www: http://www.csc.liv.ac.uk/~u5ms
--    ed.: 1.00
--
-- Mandelbrot generator with realtime exploration.
--
-- FUTURE: Real-time exploration with scrolling
--
-- SLOWED: integer.e, power(n, 2), single pixel()
-- SPEEDY: explicit ordered maths, sequential pixel()
--         no boolean if statements, explicit addition (i+i not i*2)

without type_check

include get.e
include graphics.e
include image.e
include machine.e

constant ver = 1.00             -- version

constant VGA = 19               -- 320x200x256
--constant SVGA = 256               -- 640x400x256
--constant GRAPHICS_MODE = 257    -- 640x480x256
--constant GRAPHICS_MODE = 259    -- 800x600x256

sequence vc, scanline, mpal
integer width, height
atom Xmin, Xmax, Ymin, Ymax, Imax, part

constant diff = 4               -- infinity test
constant step_y = 1             -- y plot freq

procedure init(integer gmode)
    --change graphics mode
    if graphics_mode(gmode) then
       puts(1, "Needs at least VGA graphics!\n")
       abort(1)
    end if
    vc = video_config()
    
    -- screen edge limits
    width = vc[VC_XPIXELS]
    height = vc[VC_YPIXELS]
end procedure

procedure wait()
    while get_key() = -1 do
    end while
end procedure

function abs(atom i)
    if i < 0 then
	return -i
    else
	return i
    end if
end function

function gap(atom n1, atom n2)
    return abs(n1) + abs(n2)
end function

procedure set_palette(sequence rgb)
-- set a palette
    sequence pal
    
    pal = {}
    for v = 0 to 255 do
	pal = append(pal, {v*(rgb[1]*2), v*(rgb[2]*2), v*(rgb[3]*2)})
    end for
    all_palette(pal)
end procedure

procedure zoom(atom bomb)
    atom odx, ody
    
    odx = (gap(Xmin, Xmax) / width) * bomb
    ody = (gap(Ymin, Ymax) / height) * (bomb / width * height)
    
    Xmin = Xmin - odx
    Xmax = Xmax + odx
    Ymin = Ymin - ody
    Ymax = Ymax + ody
end procedure

procedure edge(integer mx, integer my, atom d)
    atom odx, ody
    
    odx = (gap(Xmin, Xmax) / width) * d
    ody = (gap(Ymin, Ymax) / height) * d
    
    Xmin = Xmin + mx * odx
    Xmax = Xmax + mx * odx
    Ymin = Ymin + my * ody
    Ymax = Ymax + my * ody
end procedure

procedure plot(sequence origin, sequence size)
-- mandelbrot, screen section
    atom dx, dy, ca, cb, old_a, a2, b2, a, b
    integer r
    sequence colours
    --atom t
    
    scanline = repeat(0, size[1] - origin[1] - 1)
    
    dx = gap(Xmin, Xmax) / width
    dy = gap(Ymin, Ymax) / height

    --t = time()
    for y = origin[2] to size[2] - 1 by step_y do
	colours = {}
	if size[1] = width and size[2] = height then
	    part = get_key()
	    if part >= 0 then exit
		-- key to exit
	    end if
	    pixel(scanline, {origin[1], y})
	end if
	for x = origin[1] to size[1] - 1 do
	    r = 0
	    a = 0
	    b = 0
	    ca = Xmin + x * dx
	    cb = Ymin + y * dy
	    while 1 do
		old_a = a
		a2 = a*a
		b2 = b*b
		if (a2 + b2) > diff then exit
		end if
		if r > Imax then exit
		end if
		a = a2 - b2 + ca
		b = b*(old_a+old_a) + cb
		r = r + 1
	    end while
	    if r > Imax then r = 0
	    end if
	    colours = append(colours, r)
	end for
	pixel(colours, {origin[1], y})
	--Note similar
	--poke(#A0000 + (x+r)*320,p)  --*** faster, but works only in mode 19
    end for 
    --t = time() - t
    --position(1, 1)
    --printf(1, "%3.2fs\n", t)
end procedure

procedure explore()
    atom quit, k, grab

    -- set palette
    if length(mpal) = 3 then
	set_palette(mpal)
    end if
	
    quit = 0
    while 1 do

    plot({0, 0}, {width, height})
    
    while 1 do
	if part then
	    k = part
	    part = 0
	else
	    k = get_key()       -- key press
	end if
	if k != -1 then
	    if k = 13 or k = 27 then    -- RET/ESC (to menu)
		quit = 1
		exit
	    elsif k = 66 or k = 98 then     -- B, b
		mpal = {0, 0, 1}
		set_palette(mpal)
	    elsif k = 71 or k = 103 then    -- G, g
		mpal = {0, 1, 0}
		set_palette(mpal)
	    elsif k = 77 or k = 109 then    -- M, m
		mpal = {1, 1, 1}
		set_palette(mpal)
	    elsif k = 82 or k = 114 then    -- R, r
		mpal = {1, 0, 0}
		set_palette(mpal)
	    elsif k = 327 then  -- HOME (defaults)
		Xmin = -2.8
		Xmax = 2.8
		Ymin = -2.1
		Ymax = 2.1
		exit
	    elsif k = 328 then  -- UP
		edge(0, -1, 2)
		grab = save_image({0, 0}, {width, height - 2})
		plot({0, 0}, {width, 2})
		display_image({0, 2}, grab)
	    elsif k = 329 then  -- PGUP (zoom in)
		zoom(-10)
		exit
	    elsif k = 331 then  -- LEFT
		edge(-1, 0, 2)
		grab = save_image({0, 0}, {width - 2, height})
		plot({0, 0}, {2, height})
		display_image({2, 0}, grab)
	    elsif k = 333 then  -- RIGHT
		edge(1, 0, 2)
		grab = save_image({2, 0}, {width, height})
		display_image({0, 0}, grab)
		plot({width - 2, 0}, {width, height})
	    elsif k = 335 then  -- END (redraw)
		exit
	    elsif k = 336 then  -- DOWN
		edge(0, 1, 2)
		grab = save_image({0, 2}, {width, height})
		display_image({0, 0}, grab)
		plot({0, height - 2}, {width, height})
	    elsif k = 337 then  -- PGDN (zoom out)
		zoom(10)
		exit
	    end if
	end if
    end while
    if quit then
	-- revert to old graphics mode
	if graphics_mode(-1) then
	end if
	exit
    end if
    
    end while
end procedure

procedure menu()
    atom k
    
    use_vesa(1)
    
    Xmin = -2.8             -- real min (horiz)
    Xmax = 2.8              -- real max
    Ymin = -2.1             -- imag min (vert)
    Ymax = 2.1              -- imag max
    Imax = 32               -- max iterations
    mpal = {}               -- default palette
    
    while 1 do
	cursor(NO_CURSOR)
	position(1, 1)
	puts(1, "ME^2: Matt's Euphoria Mandelbrot Explorer")
	printf(1, " %1.2f\n", ver)
	puts(1, repeat('=', 46) & "\n\n")
	puts(1, "MENU\n\n")
	puts(1, "ESC\tQuit\n")
	puts(1, "RET\tMandelbrot\n")
	puts(1, "-\tDecrease Imax\n")
	puts(1, "+\tIncrease Imax\n")
	puts(1, "D\tDefault palette\n\n\n")
	puts(1, "MANDEL\n\n")
	puts(1, "CURSORS\tExplore plot\n")
	puts(1, "END\tRedraw current plot\n")
	puts(1, "HOME\tDefault values\n")
	puts(1, "PGUP\tZoom in\n")
	puts(1, "PGDN\tZoom out\n")
	puts(1, "RET/ESC\tMenu\n\n\n")
	puts(1, "BOTH\n\n")
	puts(1, "R\tRed palette\n")
	puts(1, "G\tGreen palette\n")
	puts(1, "B\tBlue palette\n")
	puts(1, "M\tMono palette\n\n\n")
	puts(1, "VALUES\n\n")
	printf(1, "Xmin = %+1.14e\tYmin = %+1.14e\n", {Xmin, Ymin})
	printf(1, "Xmax = %+1.14e\tYmax = %+1.14e\n\n", {Xmax, Ymax})
	printf(1, "Imax = %-3d\n", Imax)
	--printf(1, "Mpal = %-3d\n", length(mpal))
	k = wait_key()
	if k != -1 then         -- key pressed...
	    if k = 13 then      -- ENTER, plots
		init(VGA)
		explore()
	    elsif k = 27 then   -- ESC, quits
		exit
	    elsif k = 43 then   -- plus
		if Imax < 256 then
		    Imax = Imax + 2
		end if
	    elsif k = 45 then   -- minus
		if Imax > 2 then
		    Imax = Imax - 2
		end if
	    elsif k = 66 or k = 98 then     -- B, b
		mpal = {0, 0, 1}
	    elsif k = 68 or k = 100 then    -- D, d
		mpal = {}
	    elsif k = 71 or k = 103 then    -- G, g
		mpal = {0, 1, 0}
	    elsif k = 77 or k = 109 then    -- M, m
		mpal = {1, 1, 1}
	    elsif k = 82 or k = 114 then    -- R, r
		mpal = {1, 0, 0}
	    end if
	end if
    end while
    clear_screen()
end procedure

menu()
